@@ -57,14 +57,13 @@
 struct bftw_file {
 	/** The parent directory, if any. */
 	struct bftw_file *parent;
-	/** This file's depth in the walk. */
-	size_t depth;
-	/** The root path this file was found from. */
-	const char *root;
-
+	/** The root under which this file was found. */
+	struct bftw_file *root;
 	/** The next file in the queue, if any. */
 	struct bftw_file *next;
 
+	/** This file's depth in the walk. */
+	size_t depth;
 	/** Reference count. */
 	size_t refcount;
 	/** Index in the bftw_cache priority queue. */
@@ -311,13 +310,13 @@ static struct bftw_file *bftw_file_new(struct bftw_cache *cache, struct bftw_fil
 	file->parent = parent;
 
 	if (parent) {
-		file->depth = parent->depth + 1;
 		file->root = parent->root;
+		file->depth = parent->depth + 1;
 		file->nameoff = bftw_child_nameoff(parent);
 		bftw_file_incref(cache, parent);
 	} else {
+		file->root = file;
 		file->depth = 0;
-		file->root = name;
 		file->nameoff = 0;
 	}
 
@@ -902,7 +901,7 @@ static bool bftw_need_stat(const struct bftw_state *state) {
 	}
 
 	if (ftwbuf->typeflag == BFTW_DIR) {
-		if (state->flags & (BFTW_DETECT_CYCLES | BFTW_XDEV)) {
+		if (state->flags & (BFTW_DETECT_CYCLES | BFTW_MOUNT | BFTW_XDEV)) {
 			return true;
 		}
 #if __linux__
@@ -912,7 +911,7 @@ static bool bftw_need_stat(const struct bftw_state *state) {
 		// need to stat() to get the correct type.  We don't need to
 		// check for directories because they can only be mounted over
 		// by other directories.
-		if (bfs_maybe_mount(state->mtab, ftwbuf->path)) {
+		if (bfs_might_be_mount(state->mtab, ftwbuf->path)) {
 			return true;
 		}
 #endif
@@ -963,7 +962,7 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 
 	struct BFTW *ftwbuf = &state->ftwbuf;
 	ftwbuf->path = state->path;
-	ftwbuf->root = file ? file->root : ftwbuf->path;
+	ftwbuf->root = file ? file->root->name : ftwbuf->path;
 	ftwbuf->depth = 0;
 	ftwbuf->visit = visit;
 	ftwbuf->typeflag = BFTW_UNKNOWN;
@@ -1029,11 +1028,8 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	}
 
 	if (ftwbuf->typeflag == BFTW_DIR && (state->flags & BFTW_DETECT_CYCLES)) {
-		for (const struct bftw_file *parent = file; parent; parent = parent->parent) {
-			if (parent->depth == ftwbuf->depth) {
-				continue;
-			}
-			if (parent->dev == statbuf->dev && parent->ino == statbuf->ino) {
+		for (const struct bftw_file *ancestor = parent; ancestor; ancestor = ancestor->parent) {
+			if (ancestor->dev == statbuf->dev && ancestor->ino == statbuf->ino) {
 				ftwbuf->typeflag = BFTW_ERROR;
 				ftwbuf->error = ELOOP;
 				return;
@@ -1042,6 +1038,23 @@ static void bftw_init_ftwbuf(struct bftw_state *state, enum bftw_visit visit) {
 	}
 }
 
+/** Check if the current file is a mount point. */
+static bool bftw_is_mount(struct bftw_state *state, const char *name) {
+	const struct bftw_file *file = state->file;
+	if (!file) {
+		return false;
+	}
+
+	const struct bftw_file *parent = name ? file : file->parent;
+	if (!parent) {
+		return false;
+	}
+
+	const struct BFTW *ftwbuf = &state->ftwbuf;
+	const struct bfs_stat *statbuf = bftw_stat(ftwbuf, ftwbuf->stat_flags);
+	return statbuf && statbuf->dev != parent->dev;
+}
+
 /** Fill file identity information from an ftwbuf. */
 static void bftw_fill_id(struct bftw_file *file, const struct BFTW *ftwbuf) {
 	const struct bfs_stat *statbuf = ftwbuf->stat_cache.buf;
@@ -1072,6 +1085,10 @@ static enum bftw_action bftw_visit(struct bftw_state *state, const char *name, e
 		return BFTW_STOP;
 	}
 
+	if ((state->flags & BFTW_MOUNT) && bftw_is_mount(state, name)) {
+		return BFTW_PRUNE;
+	}
+
 	enum bftw_action ret = state->callback(ftwbuf, state->ptr);
 	switch (ret) {
 	case BFTW_CONTINUE:
@@ -1089,19 +1106,9 @@ static enum bftw_action bftw_visit(struct bftw_state *state, const char *name, e
 		goto done;
 	}
 
-	if (state->flags & BFTW_XDEV) {
-		const struct bftw_file *parent = state->file;
-		if (parent && !name) {
-			parent = parent->parent;
-		}
-
-		if (parent) {
-			const struct bfs_stat *statbuf = bftw_stat(ftwbuf, ftwbuf->stat_flags);
-			if (statbuf && statbuf->dev != parent->dev) {
-				ret = BFTW_PRUNE;
-				goto done;
-			}
-		}
+	if ((state->flags & BFTW_XDEV) && bftw_is_mount(state, name)) {
+		ret = BFTW_PRUNE;
+		goto done;
 	}
 
 done:
@@ -1206,20 +1213,32 @@ static struct bftw_reader *bftw_open(struct bftw_state *state) {
 	return reader;
 }
 
+/**
+ * Flags controlling which files get visited when releasing a reader/file.
+ */
+enum bftw_release_flags {
+	/** Don't visit anything. */
+	BFTW_VISIT_NONE = 0,
+	/** Visit the file itself. */
+	BFTW_VISIT_FILE = 1 << 0,
+	/** Visit the file's ancestors. */
+	BFTW_VISIT_PARENTS = 1 << 1,
+	/** Visit both the file and its ancestors. */
+	BFTW_VISIT_ALL = BFTW_VISIT_FILE | BFTW_VISIT_PARENTS,
+};
+
 /**
  * Close and release the reader.
  */
-static enum bftw_action bftw_release_reader(struct bftw_state *state, bool do_visit) {
+static enum bftw_action bftw_release_reader(struct bftw_state *state, enum bftw_release_flags flags) {
 	enum bftw_action ret = BFTW_CONTINUE;
 
 	struct bftw_reader *reader = &state->reader;
 	bftw_reader_close(reader);
 
 	if (reader->error != 0) {
-		if (do_visit) {
-			if (bftw_visit(state, NULL, BFTW_PRE) == BFTW_STOP) {
-				ret = BFTW_STOP;
-			}
+		if (flags & BFTW_VISIT_FILE) {
+			ret = bftw_visit(state, NULL, BFTW_PRE);
 		} else {
 			state->error = reader->error;
 		}
@@ -1232,14 +1251,13 @@ static enum bftw_action bftw_release_reader(struct bftw_state *state, bool do_vi
 /**
  * Finalize and free a file we're done with.
  */
-static enum bftw_action bftw_release_file(struct bftw_state *state, bool visit_file, bool visit_parents) {
+static enum bftw_action bftw_release_file(struct bftw_state *state, enum bftw_release_flags flags) {
 	enum bftw_action ret = BFTW_CONTINUE;
 
 	if (!(state->flags & BFTW_DEPTH)) {
-		visit_file = false;
-		visit_parents = false;
+		flags = 0;
 	}
-	bool do_visit = visit_file;
+	bool visit = flags & BFTW_VISIT_FILE;
 
 	while (state->file) {
 		if (bftw_file_decref(&state->cache, state->file) > 0) {
@@ -1247,13 +1265,11 @@ static enum bftw_action bftw_release_file(struct bftw_state *state, bool visit_f
 			break;
 		}
 
-		if (do_visit) {
-			if (bftw_visit(state, NULL, BFTW_POST) == BFTW_STOP) {
-				ret = BFTW_STOP;
-				visit_parents = false;
-			}
+		if (visit && bftw_visit(state, NULL, BFTW_POST) == BFTW_STOP) {
+			ret = BFTW_STOP;
+			flags &= ~BFTW_VISIT_PARENTS;
 		}
-		do_visit = visit_parents;
+		visit = flags & BFTW_VISIT_PARENTS;
 
 		struct bftw_file *parent = state->file->parent;
 		if (state->previous == state->file) {
@@ -1272,7 +1288,7 @@ static enum bftw_action bftw_release_file(struct bftw_state *state, bool visit_f
 static void bftw_drain_queue(struct bftw_state *state, struct bftw_queue *queue) {
 	while (queue->head) {
 		state->file = bftw_queue_pop(queue);
-		bftw_release_file(state, false, false);
+		bftw_release_file(state, BFTW_VISIT_NONE);
 	}
 }
 
@@ -1285,9 +1301,9 @@ static void bftw_drain_queue(struct bftw_state *state, struct bftw_queue *queue)
 static int bftw_state_destroy(struct bftw_state *state) {
 	dstrfree(state->path);
 
-	bftw_release_reader(state, false);
+	bftw_release_reader(state, BFTW_VISIT_NONE);
 
-	bftw_release_file(state, false, false);
+	bftw_release_file(state, BFTW_VISIT_NONE);
 	bftw_drain_queue(state, &state->prequeue);
 	bftw_drain_queue(state, &state->queue);
 
@@ -1343,10 +1359,10 @@ static int bftw_bfs(const struct bftw_args *args) {
 			}
 		}
 
-		if (bftw_release_reader(&state, true) == BFTW_STOP) {
+		if (bftw_release_reader(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
 			goto done;
 		}
-		if (bftw_release_file(&state, true, true) == BFTW_STOP) {
+		if (bftw_release_file(&state, BFTW_VISIT_ALL) == BFTW_STOP) {
 			goto done;
 		}
 	}
@@ -1371,13 +1387,13 @@ static int bftw_dfs(const struct bftw_args *args) {
 	}
 
 	while (bftw_pop(&state) > 0) {
-		bool visit_post = true;
+		enum bftw_release_flags relflags = BFTW_VISIT_ALL;
 
 		switch (bftw_visit(&state, NULL, BFTW_PRE)) {
 		case BFTW_CONTINUE:
 			break;
 		case BFTW_PRUNE:
-			visit_post = false;
+			relflags &= ~BFTW_VISIT_FILE;
 			goto next;
 		case BFTW_STOP:
 			goto done;
@@ -1391,12 +1407,12 @@ static int bftw_dfs(const struct bftw_args *args) {
 			}
 		}
 
-		if (bftw_release_reader(&state, true) == BFTW_STOP) {
+		if (bftw_release_reader(&state, relflags) == BFTW_STOP) {
 			goto done;
 		}
 
 	next:
-		if (bftw_release_file(&state, visit_post, true) == BFTW_STOP) {
+		if (bftw_release_file(&state, relflags) == BFTW_STOP) {
 			goto done;
 		}
 	}
