@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2017-2018 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2017-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -50,6 +50,7 @@
 
 static char *fake_and_arg = "-a";
 static char *fake_or_arg = "-o";
+static char *fake_not_arg = "!";
 
 /**
  * A contrained integer range.
@@ -433,6 +434,14 @@ static struct expr *optimize_and_expr(const struct opt_state *state, struct expr
 		} else if (lhs->always_false) {
 			debug_opt(state, "-O1: short-circuit: %e <==> %e\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
+		} else if (lhs->always_true && rhs == &expr_false) {
+			debug_opt(state, "-O1: strength reduction: %e <==> ", expr);
+			struct expr *ret = extract_child_expr(expr, &expr->lhs);
+			ret = negate_expr(ret, &fake_not_arg);
+			if (ret) {
+				debug_opt(state, "%e\n", ret);
+			}
+			return ret;
 		} else if (optlevel >= 2 && lhs->pure && rhs == &expr_false) {
 			debug_opt(state, "-O2: purity: %e <==> %e\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
@@ -493,6 +502,14 @@ static struct expr *optimize_or_expr(const struct opt_state *state, struct expr
 		} else if (rhs == &expr_false) {
 			debug_opt(state, "-O1: disjunctive syllogism: %e <==> %e\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
+		} else if (lhs->always_false && rhs == &expr_true) {
+			debug_opt(state, "-O1: strength reduction: %e <==> ", expr);
+			struct expr *ret = extract_child_expr(expr, &expr->lhs);
+			ret = negate_expr(ret, &fake_not_arg);
+			if (ret) {
+				debug_opt(state, "%e\n", ret);
+			}
+			return ret;
 		} else if (optlevel >= 2 && lhs->pure && rhs == &expr_true) {
 			debug_opt(state, "-O2: purity: %e <==> %e\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
@@ -578,9 +595,11 @@ static struct expr *optimize_comma_expr(const struct opt_state *state, struct ex
 		if (expr_never_returns(lhs)) {
 			debug_opt(state, "-O1: reachability: %e <==> %e\n", expr, lhs);
 			return extract_child_expr(expr, &expr->lhs);
-		}
-
-		if (optlevel >= 2 && lhs->pure) {
+		} else if ((lhs->always_true && rhs == &expr_true)
+			   || (lhs->always_false && rhs == &expr_false)) {
+			debug_opt(state, "-O1: redundancy elimination: %e <==> %e\n", expr, lhs);
+			return extract_child_expr(expr, &expr->lhs);
+		} else if (optlevel >= 2 && lhs->pure) {
 			debug_opt(state, "-O2: purity: %e <==> %e\n", expr, rhs);
 			return extract_child_expr(expr, &expr->rhs);
 		}
