@@ -23,6 +23,7 @@
 
 #include "bfs.h"
 #include "cmdline.h"
+#include "darray.h"
 #include "diag.h"
 #include "dstring.h"
 #include "eval.h"
@@ -292,7 +293,7 @@ int free_cmdline(struct cmdline *cmdline) {
 		cfclose(cerr);
 
 		free_colors(cmdline->colors);
-		free(cmdline->paths);
+		darray_free(cmdline->paths);
 		free(cmdline->argv);
 		free(cmdline);
 	}
@@ -346,6 +347,10 @@ struct parser_state {
 	const char *depth_arg;
 	/** A "-prune"-type argument if any. */
 	const char *prune_arg;
+	/** A "-mount"-type argument if any. */
+	const char *mount_arg;
+	/** An "-xdev"-type argument if any. */
+	const char *xdev_arg;
 
 	/** The current time. */
 	struct timespec now;
@@ -501,18 +506,7 @@ static char **parser_advance(struct parser_state *state, enum token_type type, s
  */
 static int parse_root(struct parser_state *state, const char *path) {
 	struct cmdline *cmdline = state->cmdline;
-	size_t i = cmdline->npaths;
-	if ((i & (i + 1)) == 0) {
-		const char **paths = realloc(cmdline->paths, 2*(i + 1)*sizeof(*paths));
-		if (!paths) {
-			return -1;
-		}
-		cmdline->paths = paths;
-	}
-
-	cmdline->paths[i] = path;
-	cmdline->npaths = i + 1;
-	return 0;
+	return DARRAY_PUSH(&cmdline->paths, &path);
 }
 
 /**
@@ -1546,10 +1540,18 @@ static struct expr *parse_ls(struct parser_state *state, int arg1, int arg2) {
 }
 
 /**
- * Parse -mount, -xdev.
+ * Parse -mount.
  */
 static struct expr *parse_mount(struct parser_state *state, int arg1, int arg2) {
+	if (state->warn) {
+		parse_warning(state,
+		              "In the future, %s will skip mount points entirely, unlike\n"
+		              "-xdev, due to http://austingroupbugs.net/view.php?id=1133.\n\n",
+		              state->argv[0]);
+	}
+
 	state->cmdline->flags |= BFTW_XDEV;
+	state->mount_arg = state->argv[0];
 	return parse_nullary_option(state);
 }
 
@@ -2408,6 +2410,15 @@ static struct expr *parse_xattr(struct parser_state *state, int arg1, int arg2)
 #endif
 }
 
+/**
+ * Parse -xdev.
+ */
+static struct expr *parse_xdev(struct parser_state *state, int arg1, int arg2) {
+	state->cmdline->flags |= BFTW_XDEV;
+	state->xdev_arg = state->argv[0];
+	return parse_nullary_option(state);
+}
+
 /**
  * Launch a pager for the help output.
  */
@@ -2594,8 +2605,8 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "  ${blu}-mindepth${rs} ${bld}N${rs}\n");
 	cfprintf(cout, "      Ignore files deeper/shallower than ${bld}N${rs}\n");
 	cfprintf(cout, "  ${blu}-mount${rs}\n");
-	cfprintf(cout, "  ${blu}-xdev${rs}\n");
-	cfprintf(cout, "      Don't descend into other mount points\n");
+	cfprintf(cout, "      Don't descend into other mount points (same as ${blu}-xdev${rs} for now, but will\n");
+	cfprintf(cout, "      skip mount points entirely in the future)\n");
 	cfprintf(cout, "  ${blu}-noleaf${rs}\n");
 	cfprintf(cout, "      Ignored; for compatibility with GNU find\n");
 	cfprintf(cout, "  ${blu}-regextype${rs} ${bld}TYPE${rs}\n");
@@ -2605,7 +2616,9 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "      Skip any files that have already been seen\n");
 	cfprintf(cout, "  ${blu}-warn${rs}\n");
 	cfprintf(cout, "  ${blu}-nowarn${rs}\n");
-	cfprintf(cout, "      Turn on or off warnings about the command line\n\n");
+	cfprintf(cout, "      Turn on or off warnings about the command line\n");
+	cfprintf(cout, "  ${blu}-xdev${rs}\n");
+	cfprintf(cout, "      Don't descend into other mount points\n\n");
 
 	cfprintf(cout, "${bld}Tests:${rs}\n\n");
 
@@ -2888,9 +2901,9 @@ static const struct table_entry parse_table[] = {
 	{"-warn", parse_warn, true},
 	{"-wholename", parse_path, false},
 	{"-writable", parse_access, W_OK},
-	{"-x", parse_mount},
+	{"-x", parse_xdev},
 	{"-xattr", parse_xattr},
-	{"-xdev", parse_mount},
+	{"-xdev", parse_xdev},
 	{"-xtype", parse_type, true},
 	{0},
 };
@@ -3183,17 +3196,23 @@ static struct expr *parse_whole_expr(struct parser_state *state) {
 		}
 	}
 
-	if (state->warn && state->depth_arg && state->prune_arg) {
-		parse_warning(state, "%s does not work in the presence of %s.\n", state->prune_arg, state->depth_arg);
-
-		if (state->interactive) {
-			fprintf(stderr, "Do you want to continue? ");
-			if (ynprompt() == 0) {
-				goto fail;
-			}
+	if (state->warn) {
+		if (state->mount_arg && state->xdev_arg) {
+			parse_warning(state, "%s is redundant in the presence of %s.\n\n", state->xdev_arg, state->mount_arg);
 		}
 
-		fprintf(stderr, "\n");
+		if (state->depth_arg && state->prune_arg) {
+			parse_warning(state, "%s does not work in the presence of %s.\n", state->prune_arg, state->depth_arg);
+
+			if (state->interactive) {
+				fprintf(stderr, "Do you want to continue? ");
+				if (ynprompt() == 0) {
+					goto fail;
+				}
+			}
+
+			fprintf(stderr, "\n");
+		}
 	}
 
 	return expr;
@@ -3255,7 +3274,7 @@ void dump_cmdline(const struct cmdline *cmdline, bool verbose) {
 		cfprintf(cerr, " ");
 	}
 
-	for (size_t i = 0; i < cmdline->npaths; ++i) {
+	for (size_t i = 0; i < darray_length(cmdline->paths); ++i) {
 		const char *path = cmdline->paths[i];
 		char c = path[0];
 		if (c == '-' || c == '(' || c == ')' || c == '!' || c == ',') {
@@ -3275,8 +3294,10 @@ void dump_cmdline(const struct cmdline *cmdline, bool verbose) {
 	if (cmdline->ignore_races) {
 		cfprintf(cerr, "${blu}-ignore_readdir_race${rs} ");
 	}
-	if (cmdline->flags & BFTW_XDEV) {
+	if (cmdline->flags & BFTW_MOUNT) {
 		cfprintf(cerr, "${blu}-mount${rs} ");
+	} else if (cmdline->flags & BFTW_XDEV) {
+		cfprintf(cerr, "${blu}-xdev{rs} ");
 	}
 	if (cmdline->mindepth != 0) {
 		cfprintf(cerr, "${blu}-mindepth${rs} ${bld}%d${rs} ", cmdline->mindepth);
@@ -3338,7 +3359,6 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 
 	cmdline->argv = NULL;
 	cmdline->paths = NULL;
-	cmdline->npaths = 0;
 	cmdline->colors = NULL;
 	cmdline->cout = NULL;
 	cmdline->cerr = NULL;
@@ -3411,6 +3431,8 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 		.last_arg = NULL,
 		.depth_arg = NULL,
 		.prune_arg = NULL,
+		.mount_arg = NULL,
+		.xdev_arg = NULL,
 	};
 
 	if (strcmp(xbasename(state.command), "find") == 0) {
@@ -3435,7 +3457,7 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 		goto fail;
 	}
 
-	if (cmdline->npaths == 0) {
+	if (darray_length(cmdline->paths) == 0) {
 		if (parse_root(&state, ".") != 0) {
 			goto fail;
 		}
