@@ -282,7 +282,9 @@ static int bfs_stat_explicit(int at_fd, const char *at_path, int at_flags, enum
 
 	if (has_statx) {
 		int ret = bfs_statx_impl(at_fd, at_path, at_flags, flags, buf);
-		if (ret != 0 && errno == ENOSYS) {
+		// EPERM is commonly returned in a seccomp() sandbox that does
+		// not allow statx()
+		if (ret != 0 && (errno == ENOSYS || errno == EPERM)) {
 			has_statx = false;
 		} else {
 			return ret;
@@ -299,6 +301,12 @@ int bfs_stat(int at_fd, const char *at_path, enum bfs_stat_flag flags, struct bf
 		at_flags |= AT_SYMLINK_NOFOLLOW;
 	}
 
+#ifdef AT_STATX_DONT_SYNC
+	if (flags & BFS_STAT_NOSYNC) {
+		at_flags |= AT_STATX_DONT_SYNC;
+	}
+#endif
+
 	if (at_path) {
 		return bfs_stat_explicit(at_fd, at_path, at_flags, flags, buf);
 	}
