@@ -2,7 +2,7 @@
 
 ############################################################################
 # bfs                                                                      #
-# Copyright (C) 2015-2018 Tavian Barnes <tavianator@tavianator.com>        #
+# Copyright (C) 2015-2020 Tavian Barnes <tavianator@tavianator.com>        #
 #                                                                          #
 # Permission to use, copy, modify, and/or distribute this software for any #
 # purpose with or without fee is hereby granted.                           #
@@ -16,12 +16,65 @@
 # OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.           #
 ############################################################################
 
+set -e
 set -o physical
 umask 022
 
 export LC_ALL=C
 export TZ=UTC
 
+if [ -t 1 ]; then
+    BLD="$(printf '\033[01m')"
+    RED="$(printf '\033[01;31m')"
+    GRN="$(printf '\033[01;32m')"
+    YLW="$(printf '\033[01;33m')"
+    BLU="$(printf '\033[01;34m')"
+    MAG="$(printf '\033[01;35m')"
+    CYN="$(printf '\033[01;36m')"
+    RST="$(printf '\033[0m')"
+fi
+
+if [ "$EUID" -eq 0 ]; then
+    cat >&2 <<EOF
+${RED}error:${RST} These tests expect filesystem permissions to be enforced, and therefore
+will not work when run as ${BLD}$(id -un)${RST}.
+EOF
+    exit 1
+fi
+
+function usage() {
+    local pad=$(printf "%*s" ${#0} "")
+    cat <<EOF
+Usage: ${GRN}$0${RST} [${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}] [${BLU}--posix${RST}] [${BLU}--bsd${RST}] [${BLU}--gnu${RST}] [${BLU}--all${RST}] [${BLU}--sudo${RST}]
+       $pad [${BLU}--noclean${RST}] [${BLU}--update${RST}] [${BLU}--verbose${RST}] [${BLU}--help${RST}]
+       $pad [${BLD}test_*${RST} [${BLD}test_*${RST} ...]]
+
+  ${BLU}--bfs${RST}=${MAG}path/to/bfs${RST}
+      Set the path to the bfs executable to test (default: ${MAG}./bfs${RST})
+
+  ${BLU}--posix${RST}, ${BLU}--bsd${RST}, ${BLU}--gnu${RST}, ${BLU}--all${RST}
+      Choose which test cases to run (default: ${BLU}--all${RST})
+
+  ${BLU}--sudo${RST}
+      Run tests that require root (not included in ${BLU}--all${RST})
+
+  ${BLU}--noclean${RST}
+      Keep the test directories around after the run
+
+  ${BLU}--update${RST}
+      Update the expected outputs for the test cases
+
+  ${BLU}--verbose${RST}
+      Log the commands that get executed
+
+  ${BLU}--help${RST}
+      This message
+
+  ${BLD}test_*${RST}
+      Select individual test cases to run
+EOF
+}
+
 function _realpath() {
     (
         cd "${1%/*}"
@@ -31,179 +84,75 @@ function _realpath() {
 
 BFS="$(_realpath ./bfs)"
 TESTS="$(_realpath ./tests)"
+UNAME="$(uname)"
 
-# The temporary directory that will hold our test data
-TMP="$(mktemp -d "${TMPDIR:-/tmp}"/bfs.XXXXXXXXXX)"
-chown "$(id -u):$(id -g)" "$TMP"
-
-# Clean up temporary directories on exit
+DEFAULT=yes
+POSIX=
+BSD=
+GNU=
+ALL=
+SUDO=
 CLEAN=yes
-function cleanup() {
-    if [ ! "$CLEAN" ]; then
-        return
-    fi
+UPDATE=
+VERBOSE=
+EXPLICIT=
 
-    # Don't force rm to deal with long paths
-    for dir in "$TMP"/deep/*/*; do
-        if [ -d "$dir" ]; then
-            (cd "$dir" && rm -rf *)
-        fi
-    done
+enabled_tests=()
 
-    rm -rf "$TMP"
-}
-trap cleanup EXIT
-
-# Install a file, creating any parent directories
-function installp() {
-    local target="${@: -1}"
-    mkdir -p "${target%/*}"
-    install "$@"
-}
-
-# Like a mythical touch -p
-function touchp() {
-    for arg; do
-        installp -m644 /dev/null "$arg"
-    done
-}
-
-# Creates a simple file+directory structure for tests
-function make_basic() {
-    touchp "$1/a"
-    touchp "$1/b"
-    touchp "$1/c/d"
-    touchp "$1/e/f"
-    mkdir -p "$1/g/h"
-    mkdir -p "$1/i"
-    touchp "$1/j/foo"
-    touchp "$1/k/foo/bar"
-    touchp "$1/l/foo/bar/baz"
-    echo baz >"$1/l/foo/bar/baz"
-}
-make_basic "$TMP/basic"
-
-# Creates a file+directory structure with various permissions for tests
-function make_perms() {
-    installp -m000 /dev/null "$1/0"
-    installp -m444 /dev/null "$1/r"
-    installp -m222 /dev/null "$1/w"
-    installp -m644 /dev/null "$1/rw"
-    installp -m555 /dev/null "$1/rx"
-    installp -m311 /dev/null "$1/wx"
-    installp -m755 /dev/null "$1/rwx"
-}
-make_perms "$TMP/perms"
-
-# Creates a file+directory structure with various symbolic and hard links
-function make_links() {
-    touchp "$1/file"
-    ln -s file "$1/symlink"
-    ln "$1/file" "$1/hardlink"
-    ln -s nowhere "$1/broken"
-    ln -s symlink/file "$1/notdir"
-    mkdir -p "$1/deeply/nested/dir"
-    touchp "$1/deeply/nested/file"
-    ln -s file "$1/deeply/nested/link"
-    ln -s deeply/nested "$1/skip"
-}
-make_links "$TMP/links"
-
-# Creates a file+directory structure with symbolic link loops
-function make_loops() {
-    touchp "$1/file"
-    ln -s file "$1/symlink"
-    ln -s nowhere "$1/broken"
-    ln -s symlink/file "$1/notdir"
-    ln -s loop "$1/loop"
-    mkdir -p "$1/deeply/nested/dir"
-    ln -s ../../deeply "$1/deeply/nested/loop"
-    ln -s deeply/nested/loop/nested "$1/skip"
-}
-make_loops "$TMP/loops"
-
-# Creates a file+directory structure with varying timestamps
-function make_times() {
-    mkdir -p "$1"
-    touch -t 199112140000 "$1/a"
-    touch -t 199112140001 "$1/b"
-    touch -t 199112140002 "$1/c"
-    ln -s a "$1/l"
-    touch -h -t 199112140003 "$1/l"
-    touch -t 199112140004 "$1"
-}
-make_times "$TMP/times"
-
-# Creates a file+directory structure with various weird file/directory names
-function make_weirdnames() {
-    touchp "$1/-/a"
-    touchp "$1/(/b"
-    touchp "$1/(-/c"
-    touchp "$1/!/d"
-    touchp "$1/!-/e"
-    touchp "$1/,/f"
-    touchp "$1/)/g"
-    touchp "$1/.../h"
-    touchp "$1/\\/i"
-    touchp "$1/ /j"
-}
-make_weirdnames "$TMP/weirdnames"
-
-# Creates a very deep directory structure for testing PATH_MAX handling
-function make_deep() {
-    mkdir -p "$1"
-
-    # $name will be 255 characters, aka _XOPEN_NAME_MAX
-    local name="0123456789ABCDEF"
-    name="${name}${name}${name}${name}"
-    name="${name}${name}${name}${name}"
-    name="${name:0:255}"
-
-    for i in {0..9} A B C D E F; do
-        (
-            mkdir "$1/$i"
-            cd "$1/$i"
-
-            # 16 * 256 == 4096 == PATH_MAX
-            for j in {1..16}; do
-                mkdir "$name"
-                cd "$name" 2>/dev/null
-            done
-
-            touch "$name"
-        )
-    done
-}
-make_deep "$TMP/deep"
-
-# Creates a directory structure with many different types, and therefore colors
-function make_rainbow() {
-    touchp "$1/file.txt"
-    touchp "$1/file.dat"
-    ln -s file.txt "$1/link.txt"
-    touchp "$1/mh1"
-    ln "$1/mh1" "$1/mh2"
-    mkfifo "$1/pipe"
-    # TODO: block
-    # TODO: chardev
-    ln -s nowhere "$1/broken"
-    "$TESTS/mksock" "$1/socket"
-    touchp "$1"/s{u,g,ug}id
-    chmod u+s "$1"/su{,g}id
-    chmod g+s "$1"/s{u,}gid
-    mkdir "$1/ow" "$1"/sticky{,_ow}
-    chmod o+w "$1"/*ow
-    chmod +t "$1"/sticky*
-    touchp "$1"/exec.sh
-    chmod +x "$1"/exec.sh
-}
-make_rainbow "$TMP/rainbow"
-
-# Creates a scratch directory that tests can modify
-function make_scratch() {
-    mkdir -p "$1"
-}
-make_scratch "$TMP/scratch"
+for arg; do
+    case "$arg" in
+        --bfs=*)
+            BFS="${arg#*=}"
+            ;;
+        --posix)
+            DEFAULT=
+            POSIX=yes
+            ;;
+        --bsd)
+            DEFAULT=
+            POSIX=yes
+            BSD=yes
+            ;;
+        --gnu)
+            DEFAULT=
+            POSIX=yes
+            GNU=yes
+            ;;
+        --all)
+            DEFAULT=
+            POSIX=yes
+            BSD=yes
+            GNU=yes
+            ALL=yes
+            ;;
+        --sudo)
+            DEFAULT=
+            SUDO=yes
+            ;;
+        --noclean)
+            CLEAN=
+            ;;
+        --update)
+            UPDATE=yes
+            ;;
+        --verbose)
+            VERBOSE=yes
+            ;;
+        --help)
+            usage
+            exit 0
+            ;;
+        test_*)
+            EXPLICIT=yes
+            enabled_tests+=("$arg")
+            ;;
+        *)
+            printf "${RED}error:${RST} Unrecognized option '%s'.\n\n" "$arg" >&2
+            usage >&2
+            exit 1
+            ;;
+    esac
+done
 
 posix_tests=(
     # General parsing
@@ -328,6 +277,9 @@ bsd_tests=(
 
     # Primaries
 
+    test_acl
+    test_L_acl
+
     test_delete
 
     test_depth_maxdepth_1
@@ -350,6 +302,7 @@ bsd_tests=(
     test_execdir_slash
     test_execdir_slash_pwd
     test_execdir_slashes
+    test_execdir_ulimit
 
     test_exit
 
@@ -484,6 +437,7 @@ gnu_tests=(
     test_execdir_slash
     test_execdir_slash_pwd
     test_execdir_slashes
+    test_execdir_ulimit
 
     test_executable
 
@@ -499,7 +453,9 @@ gnu_tests=(
 
     test_gid
     test_gid_plus
+    test_gid_plus_plus
     test_gid_minus
+    test_gid_minus_plus
 
     test_ignore_readdir_race
     test_ignore_readdir_race_root
@@ -551,6 +507,7 @@ gnu_tests=(
     test_print_error
 
     test_printf
+    test_printf_empty
     test_printf_slash
     test_printf_slashes
     test_printf_trailing_slash
@@ -562,6 +519,7 @@ gnu_tests=(
     test_printf_leak
     test_printf_nul
     test_printf_Y_error
+    test_printf_H
 
     test_quit
     test_quit_child
@@ -595,7 +553,9 @@ gnu_tests=(
 
     test_uid
     test_uid_plus
+    test_uid_plus_plus
     test_uid_minus
+    test_uid_minus_plus
 
     test_writable
 
@@ -608,6 +568,9 @@ gnu_tests=(
     test_and_purity
     test_not_reachability
     test_comma_reachability
+    test_and_false_or_true
+    test_comma_redundant_true
+    test_comma_redundant_false
 )
 
 bfs_tests=(
@@ -621,19 +584,44 @@ bfs_tests=(
     # Primaries
 
     test_color
+    test_color_L
+    test_color_rs_lc_rc_ec
+    test_color_escapes
+    test_color_nul
+    test_color_ln_target
+    test_color_L_ln_target
     test_color_mh
     test_color_mh0
     test_color_or
     test_color_mi
     test_color_or_mi
-    test_color_or0_mi
     test_color_or_mi0
+    test_color_or0_mi
+    test_color_or0_mi0
+    test_color_su_sg0
+    test_color_su0_sg
+    test_color_su0_sg0
+    test_color_st_tw_ow0
+    test_color_st_tw0_ow
+    test_color_st_tw0_ow0
+    test_color_st0_tw_ow
+    test_color_st0_tw_ow0
+    test_color_st0_tw0_ow
+    test_color_st0_tw0_ow0
     test_color_ext
     test_color_ext0
+    test_color_ext_override
+    test_color_ext_underride
     test_color_missing_colon
+    test_color_no_stat
+    test_color_L_no_stat
+    test_color_star
+    test_color_ls
 
     test_execdir_plus
 
+    test_help
+
     test_hidden
 
     test_nohidden
@@ -647,6 +635,12 @@ bfs_tests=(
 
     test_type_multi
 
+    test_unique
+    test_unique_depth
+    test_L_unique
+    test_L_unique_loops
+    test_L_unique_depth
+
     test_xtype_multi
     test_xtype_reorder
 
@@ -654,111 +648,235 @@ bfs_tests=(
     test_deep_strict
 )
 
-function usage() {
-    local pad=$(printf "%*s" ${#0} "")
-    cat <<EOF
-Usage: $0 [--bfs=path/to/bfs] [--posix|--bsd|--gnu|--all]
-       $pad [--noclean] [--update] [--verbose] [--help]
-       $pad [test_* [test_* ...]]
+sudo_tests=(
+    test_capable
+    test_L_capable
 
-  --bfs=path/to/bfs
-      Set the path to the bfs executable to test (default: ./bfs)
+    test_mount
+    test_L_mount
+    test_xdev
+    test_L_xdev
 
-  --posix|--bsd|--gnu|--all
-      Restrict the set of test cases to run
+    test_inum_mount
+    test_inum_bind_mount
+    test_type_bind_mount
+    test_xtype_bind_mount
+)
 
-  --noclean
-      Keep the test directories around after the run
+if [ "$UNAME" = "Linux" ]; then
+    sudo_tests+=(
+        test_xattr
+        test_L_xattr
+    )
+else
+    bsd_tests+=(
+        test_xattr
+        test_L_xattr
+    )
+fi
 
-  --update
-      Update the expected outputs for the test cases
-
-  --verbose
-      Log the commands that get executed
-
-  --help
-      This message
-
-  test_*
-      Select individual test cases to run
-EOF
-}
-
-BSD=yes
-GNU=yes
-ALL=yes
-
-function enable_tests() {
-    for test; do
-        eval run_$test=yes
-    done
-}
-
-for arg; do
-    case "$arg" in
-        --bfs=*)
-            BFS="${arg#*=}"
-            ;;
-        --posix)
-            BSD=
-            GNU=
-            ALL=
-            ;;
-        --bsd)
-            BSD=yes
-            GNU=
-            ALL=
-            ;;
-        --gnu)
-            BSD=
-            GNU=yes
-            ALL=
-            ;;
-        --all)
-            BSD=yes
-            GNU=yes
-            ALL=yes
-            ;;
-        --noclean)
-            CLEAN=
-            ;;
-        --update)
-            UPDATE=yes
-            ;;
-        --verbose)
-            VERBOSE=yes
-            ;;
-        --help)
-            usage
-            exit 0
-            ;;
-        test_*)
-            EXPLICIT=yes
-            enable_tests "$arg"
-            ;;
-        *)
-            printf "Unrecognized option '%s'.\n\n" "$arg" >&2
-            usage >&2
-            exit 1
-            ;;
-    esac
-done
-
-if [ ! "$CLEAN" ]; then
-    echo "Test files saved to $TMP"
+if [ "$DEFAULT" ]; then
+    POSIX=yes
+    BSD=yes
+    GNU=yes
+    ALL=yes
 fi
 
 if [ ! "$EXPLICIT" ]; then
-    enable_tests "${posix_tests[@]}"
-    [ "$BSD" ] && enable_tests "${bsd_tests[@]}"
-    [ "$GNU" ] && enable_tests "${gnu_tests[@]}"
-    [ "$ALL" ] && enable_tests "${bfs_tests[@]}"
+    [ "$POSIX" ] && enabled_tests+=("${posix_tests[@]}")
+    [ "$BSD" ] && enabled_tests+=("${bsd_tests[@]}")
+    [ "$GNU" ] && enabled_tests+=("${gnu_tests[@]}")
+    [ "$ALL" ] && enabled_tests+=("${bfs_tests[@]}")
+    [ "$SUDO" ] && enabled_tests+=("${sudo_tests[@]}")
 fi
 
+eval enabled_tests=($(printf '%q\n' "${enabled_tests[@]}" | sort -u))
+
+# The temporary directory that will hold our test data
+TMP="$(mktemp -d "${TMPDIR:-/tmp}"/bfs.XXXXXXXXXX)"
+chown "$(id -u):$(id -g)" "$TMP"
+
+# Clean up temporary directories on exit
+function cleanup() {
+    # Don't force rm to deal with long paths
+    for dir in "$TMP"/deep/*/*; do
+        if [ -d "$dir" ]; then
+            (cd "$dir" && rm -rf *)
+        fi
+    done
+
+    # In case a test left anything weird in scratch/
+    if [ -e "$TMP"/scratch ]; then
+        chmod -R +rX "$TMP"/scratch
+    fi
+
+    rm -rf "$TMP"
+}
+
+if [ "$CLEAN" ]; then
+    trap cleanup EXIT
+else
+    echo "Test files saved to $TMP"
+fi
+
+# Install a file, creating any parent directories
+function installp() {
+    local target="${@: -1}"
+    mkdir -p "${target%/*}"
+    install "$@"
+}
+
+# Like a mythical touch -p
+function touchp() {
+    for arg; do
+        installp -m644 /dev/null "$arg"
+    done
+}
+
+# Creates a simple file+directory structure for tests
+function make_basic() {
+    touchp "$1/a"
+    touchp "$1/b"
+    touchp "$1/c/d"
+    touchp "$1/e/f"
+    mkdir -p "$1/g/h"
+    mkdir -p "$1/i"
+    touchp "$1/j/foo"
+    touchp "$1/k/foo/bar"
+    touchp "$1/l/foo/bar/baz"
+    echo baz >"$1/l/foo/bar/baz"
+}
+make_basic "$TMP/basic"
+
+# Creates a file+directory structure with various permissions for tests
+function make_perms() {
+    installp -m000 /dev/null "$1/0"
+    installp -m444 /dev/null "$1/r"
+    installp -m222 /dev/null "$1/w"
+    installp -m644 /dev/null "$1/rw"
+    installp -m555 /dev/null "$1/rx"
+    installp -m311 /dev/null "$1/wx"
+    installp -m755 /dev/null "$1/rwx"
+}
+make_perms "$TMP/perms"
+
+# Creates a file+directory structure with various symbolic and hard links
+function make_links() {
+    touchp "$1/file"
+    ln -s file "$1/symlink"
+    ln "$1/file" "$1/hardlink"
+    ln -s nowhere "$1/broken"
+    ln -s symlink/file "$1/notdir"
+    mkdir -p "$1/deeply/nested/dir"
+    touchp "$1/deeply/nested/file"
+    ln -s file "$1/deeply/nested/link"
+    ln -s deeply/nested "$1/skip"
+}
+make_links "$TMP/links"
+
+# Creates a file+directory structure with symbolic link loops
+function make_loops() {
+    touchp "$1/file"
+    ln -s file "$1/symlink"
+    ln -s nowhere "$1/broken"
+    ln -s symlink/file "$1/notdir"
+    ln -s loop "$1/loop"
+    mkdir -p "$1/deeply/nested/dir"
+    ln -s ../../deeply "$1/deeply/nested/loop"
+    ln -s deeply/nested/loop/nested "$1/skip"
+}
+make_loops "$TMP/loops"
+
+# Creates a file+directory structure with varying timestamps
+function make_times() {
+    mkdir -p "$1"
+    touch -t 199112140000 "$1/a"
+    touch -t 199112140001 "$1/b"
+    touch -t 199112140002 "$1/c"
+    ln -s a "$1/l"
+    touch -h -t 199112140003 "$1/l"
+    touch -t 199112140004 "$1"
+}
+make_times "$TMP/times"
+
+# Creates a file+directory structure with various weird file/directory names
+function make_weirdnames() {
+    touchp "$1/-/a"
+    touchp "$1/(/b"
+    touchp "$1/(-/c"
+    touchp "$1/!/d"
+    touchp "$1/!-/e"
+    touchp "$1/,/f"
+    touchp "$1/)/g"
+    touchp "$1/.../h"
+    touchp "$1/\\/i"
+    touchp "$1/ /j"
+}
+make_weirdnames "$TMP/weirdnames"
+
+# Creates a very deep directory structure for testing PATH_MAX handling
+function make_deep() {
+    mkdir -p "$1"
+
+    # $name will be 255 characters, aka _XOPEN_NAME_MAX
+    local name="0123456789ABCDEF"
+    name="${name}${name}${name}${name}"
+    name="${name}${name}${name}${name}"
+    name="${name:0:255}"
+
+    for i in {0..9} A B C D E F; do
+        (
+            mkdir "$1/$i"
+            cd "$1/$i"
+
+            # 16 * 256 == 4096 == PATH_MAX
+            for j in {1..16}; do
+                mkdir "$name"
+                cd "$name" 2>/dev/null
+            done
+
+            touch "$name"
+        )
+    done
+}
+make_deep "$TMP/deep"
+
+# Creates a directory structure with many different types, and therefore colors
+function make_rainbow() {
+    touchp "$1/file.txt"
+    touchp "$1/file.dat"
+    touchp "$1/star".{gz,tar,tar.gz}
+    ln -s file.txt "$1/link.txt"
+    touchp "$1/mh1"
+    ln "$1/mh1" "$1/mh2"
+    mkfifo "$1/pipe"
+    # TODO: block
+    ln -s /dev/null "$1/chardev_link"
+    ln -s nowhere "$1/broken"
+    "$TESTS/mksock" "$1/socket"
+    touchp "$1"/s{u,g,ug}id
+    chmod u+s "$1"/su{,g}id
+    chmod g+s "$1"/s{u,}gid
+    mkdir "$1/ow" "$1"/sticky{,_ow}
+    chmod o+w "$1"/*ow
+    chmod +t "$1"/sticky*
+    touchp "$1"/exec.sh
+    chmod +x "$1"/exec.sh
+}
+make_rainbow "$TMP/rainbow"
+
+# Creates a scratch directory that tests can modify
+function make_scratch() {
+    mkdir -p "$1"
+}
+make_scratch "$TMP/scratch"
+
 function bfs_sort() {
     awk -F/ '{ print NF - 1 " " $0 }' | sort -n | cut -d' ' -f2-
 }
 
+# Close stdin so bfs doesn't think we're interactive
+exec </dev/null
 
 if [ "$VERBOSE" ]; then
     # dup stdout for verbose logging even when redirected
@@ -768,24 +886,24 @@ fi
 function bfs_verbose() {
     if [ "$VERBOSE" ]; then
         if [ -t 3 ]; then
-            printf '\033[1;32m%q\033[0m ' $BFS >&3
+            printf "${GRN}%q${RST} " $BFS >&3
 
             local expr_started=
             for arg; do
                 if [[ $arg == -[A-Z]* ]]; then
-                    printf '\033[1;36m%q\033[0m ' "$arg" >&3
+                    printf "${CYN}%q${RST} " "$arg" >&3
                 elif [[ $arg == [\(!] || $arg == -[ao] || $arg == -and || $arg == -or || $arg == -not ]]; then
                     expr_started=yes
-                    printf '\033[1;31m%q\033[0m ' "$arg" >&3
+                    printf "${RED}%q${RST} " "$arg" >&3
                 elif [[ $expr_started && $arg == [\),] ]]; then
-                    printf '\033[1;31m%q\033[0m ' "$arg" >&3
+                    printf "${RED}%q${RST} " "$arg" >&3
                 elif [[ $arg == -?* ]]; then
                     expr_started=yes
-                    printf '\033[1;34m%q\033[0m ' "$arg" >&3
+                    printf "${BLU}%q${RST} " "$arg" >&3
                 elif [ "$expr_started" ]; then
-                    printf '\033[1m%q\033[0m ' "$arg" >&3
+                    printf "${BLD}%q${RST} " "$arg" >&3
                 else
-                    printf '\033[1;35m%q\033[0m ' "$arg" >&3
+                    printf "${MAG}%q${RST} " "$arg" >&3
                 fi
             done
         else
@@ -852,7 +970,13 @@ function closefrom() {
     done
 }
 
+function inum() {
+    ls -id "$@" | awk '{ print $1 }'
+}
+
+
 cd "$TMP"
+set +e
 
 # Test cases
 
@@ -935,11 +1059,11 @@ function test_name_trailing_slash() {
 }
 
 function test_name_slash() {
-    bfs_diff / -maxdepth 0 -name / 2>/dev/null
+    bfs_diff / -maxdepth 0 -name /
 }
 
 function test_name_slashes() {
-    bfs_diff /// -maxdepth 0 -name / 2>/dev/null
+    bfs_diff /// -maxdepth 0 -name /
 }
 
 function test_path() {
@@ -978,10 +1102,18 @@ function test_gid_plus() {
     bfs_diff basic -gid +0
 }
 
+function test_gid_plus_plus() {
+    bfs_diff basic -gid +0
+}
+
 function test_gid_minus() {
     bfs_diff basic -gid "-$(($(id -g) + 1))"
 }
 
+function test_gid_minus_plus() {
+    bfs_diff basic -gid "-+$(($(id -g) + 1))"
+}
+
 function test_uid() {
     bfs_diff basic -uid "$(id -u)"
 }
@@ -990,10 +1122,18 @@ function test_uid_plus() {
     bfs_diff basic -uid +0
 }
 
+function test_uid_plus_plus() {
+    bfs_diff basic -uid ++0
+}
+
 function test_uid_minus() {
     bfs_diff basic -uid "-$(($(id -u) + 1))"
 }
 
+function test_uid_minus_plus() {
+    bfs_diff basic -uid "-+$(($(id -u) + 1))"
+}
+
 function test_newer() {
     bfs_diff times -newer times/a
 }
@@ -1063,7 +1203,7 @@ function test_L_notdir() {
 }
 
 function test_L_loops() {
-    # POSIX says it's okay to either stop of keep going on seeing a filesystem
+    # POSIX says it's okay to either stop or keep going on seeing a filesystem
     # loop, as long as a diagnostic is printed
     local errors="$(invoke_bfs -L loops 2>&1 >/dev/null)"
     [ -n "$errors" ]
@@ -1150,8 +1290,8 @@ function test_xtype_multi() {
 function test_xtype_reorder() {
     # Make sure -xtype is not reordered in front of anything -- if -xtype runs
     # before -links 100, it will report an ELOOP error
-    bfs_diff links -links 100 -xtype l
-    invoke_bfs links -links 100 -xtype l
+    bfs_diff loops -links 100 -xtype l
+    invoke_bfs loops -links 100 -xtype l
 }
 
 function test_iname() {
@@ -1171,11 +1311,11 @@ function test_ilname() {
 }
 
 function test_L_lname() {
-    bfs_diff -L links -lname '[aq]' 2>/dev/null
+    bfs_diff -L links -lname '[aq]'
 }
 
 function test_L_ilname() {
-    bfs_diff -L links -ilname '[AQ]' 2>/dev/null
+    bfs_diff -L links -ilname '[AQ]'
 }
 
 function test_user_name() {
@@ -1259,7 +1399,9 @@ function test_execdir() {
 }
 
 function test_execdir_plus() {
-    bfs_diff basic -execdir "$TESTS/sort-args.sh" '{}' +
+    if [[ "$BFS" != *"-S dfs"* ]]; then
+        bfs_diff basic -execdir "$TESTS/sort-args.sh" '{}' +
+    fi
 }
 
 function test_execdir_substring() {
@@ -1289,6 +1431,16 @@ function test_execdir_slashes() {
     bfs_diff /// -maxdepth 0 -execdir echo '{}' ';'
 }
 
+function test_execdir_ulimit() {
+    rm -rf scratch/*
+    mkdir -p scratch/a/b/c/d/e/f/g/h/i/j/k/l/m/n/o/p/q/r/s/t/u/v/w/x/y/z
+    mkdir -p scratch/a/b/c/d/e/f/g/h/i/j/k/l/m/0/1/2/3/4/5/6/7/8/9/A/B/C
+
+    closefrom 4
+    ulimit -n 13
+    bfs_diff scratch -execdir echo '{}' ';'
+}
+
 function test_weird_names() {
     cd weirdnames
     bfs_diff '-' '(-' '!-' ',' ')' './(' './!' \( \! -print -o -print \)
@@ -1649,8 +1801,7 @@ function test_quit_implicit_print() {
 }
 
 function test_inum() {
-    local inode="$(ls -id basic/k/foo/bar | awk '{ print $1 }')"
-    bfs_diff basic -inum "$inode"
+    bfs_diff basic -inum "$(inum basic/k/foo/bar)"
 }
 
 function test_nogroup() {
@@ -1677,6 +1828,10 @@ function test_printf() {
     bfs_diff basic -printf '%%p(%p) %%d(%d) %%f(%f) %%h(%h) %%H(%H) %%P(%P) %%m(%m) %%M(%M) %%y(%y)\n'
 }
 
+function test_printf_empty() {
+    bfs_diff basic -printf ''
+}
+
 function test_printf_slash() {
     bfs_diff / -maxdepth 0 -printf '(%h)/(%f)\n'
 }
@@ -1750,6 +1905,10 @@ function test_printf_Y_error() {
     [ $ret -eq $EX_BFS ]
 }
 
+function test_printf_H() {
+    bfs_diff basic links -printf '%%p(%p) %%d(%d) %%f(%f) %%h(%h) %%H(%H) %%P(%P) %%y(%y)\n'
+}
+
 function test_fstype() {
     fstype="$(invoke_bfs basic -maxdepth 0 -printf '%F\n')"
     bfs_diff basic -fstype "$fstype"
@@ -1819,6 +1978,41 @@ function test_color() {
     LS_COLORS= bfs_diff rainbow -color
 }
 
+function test_color_L() {
+    LS_COLORS= bfs_diff -L rainbow -color
+}
+
+function test_color_rs_lc_rc_ec() {
+    LS_COLORS="rs=RS:lc=LC:rc=RC:ec=EC:" bfs_diff rainbow -color
+}
+
+function test_color_escapes() {
+    LS_COLORS="lc=\e[:rc=\155\::ec=^[\x5B\x6d:" bfs_diff rainbow -color
+}
+
+function test_color_nul() {
+    local EXPECTED="$TESTS/${FUNCNAME[0]}.out"
+    if [ "$UPDATE" ]; then
+        local ACTUAL="$EXPECTED"
+    else
+        local ACTUAL="$TMP/${FUNCNAME[0]}.out"
+    fi
+
+    LS_COLORS="ec=\33[m\0:" invoke_bfs rainbow -color -maxdepth 0 >"$ACTUAL"
+
+    if [ ! "$UPDATE" ]; then
+        diff -u "$EXPECTED" "$ACTUAL"
+    fi
+}
+
+function test_color_ln_target() {
+    LS_COLORS="ln=target:or=01;31:mi=01;33:" bfs_diff rainbow -color
+}
+
+function test_color_L_ln_target() {
+    LS_COLORS="ln=target:or=01;31:mi=01;33:" bfs_diff -L rainbow -color
+}
+
 function test_color_mh() {
     LS_COLORS="mh=01:" bfs_diff rainbow -color
 }
@@ -1839,12 +2033,56 @@ function test_color_or_mi() {
     LS_COLORS="or=01;31:mi=01;33:" bfs_diff rainbow -color
 }
 
+function test_color_or_mi0() {
+    LS_COLORS="or=01;31:mi=00:" bfs_diff rainbow -color
+}
+
 function test_color_or0_mi() {
     LS_COLORS="or=00:mi=01;33:" bfs_diff rainbow -color
 }
 
-function test_color_or_mi0() {
-    LS_COLORS="or=01;31:mi=00:" bfs_diff rainbow -color
+function test_color_or0_mi0() {
+    LS_COLORS="or=00:mi=00:" bfs_diff rainbow -color
+}
+
+function test_color_su_sg0() {
+    LS_COLORS="su=37;41:sg=00:" bfs_diff rainbow -color
+}
+
+function test_color_su0_sg() {
+    LS_COLORS="su=00:sg=30;43:" bfs_diff rainbow -color
+}
+
+function test_color_su0_sg0() {
+    LS_COLORS="su=00:sg=00:" bfs_diff rainbow -color
+}
+
+function test_color_st_tw_ow0() {
+    LS_COLORS="st=37;44:tw=40;32:ow=00:" bfs_diff rainbow -color
+}
+
+function test_color_st_tw0_ow() {
+    LS_COLORS="st=37;44:tw=00:ow=34;42:" bfs_diff rainbow -color
+}
+
+function test_color_st_tw0_ow0() {
+    LS_COLORS="st=37;44:tw=00:ow=00:" bfs_diff rainbow -color
+}
+
+function test_color_st0_tw_ow() {
+    LS_COLORS="st=00:tw=40;32:ow=34;42:" bfs_diff rainbow -color
+}
+
+function test_color_st0_tw_ow0() {
+    LS_COLORS="st=00:tw=40;32:ow=00:" bfs_diff rainbow -color
+}
+
+function test_color_st0_tw0_ow() {
+    LS_COLORS="st=00:tw=00:ow=34;42:" bfs_diff rainbow -color
+}
+
+function test_color_st0_tw0_ow0() {
+    LS_COLORS="st=00:tw=00:ow=00:" bfs_diff rainbow -color
 }
 
 function test_color_ext() {
@@ -1855,10 +2093,58 @@ function test_color_ext0() {
     LS_COLORS="*.txt=00:" bfs_diff rainbow -color
 }
 
+function test_color_ext_override() {
+    LS_COLORS="*.tar.gz=01;31:*.tar=01;32:*.gz=01;33:" bfs_diff rainbow -color
+}
+
+function test_color_ext_underride() {
+    LS_COLORS="*.gz=01;33:*.tar=01;32:*.tar.gz=01;31:" bfs_diff rainbow -color
+}
+
 function test_color_missing_colon() {
     LS_COLORS="*.txt=01" bfs_diff rainbow -color
 }
 
+function test_color_no_stat() {
+    LS_COLORS="mh=0:ex=0:sg=0:su=0:st=0:ow=0:tw=0:*.txt=01:" bfs_diff rainbow -color
+}
+
+function test_color_L_no_stat() {
+    LS_COLORS="mh=0:ex=0:sg=0:su=0:st=0:ow=0:tw=0:*.txt=01:" bfs_diff -L rainbow -color
+}
+
+function test_color_star() {
+    # Regression test: don't segfault on LS_COLORS="*"
+    LS_COLORS="*" bfs_diff rainbow -color
+}
+
+function test_color_ls() {
+    rm -rf scratch/*
+    touchp scratch/foo/bar/baz
+    ln -s foo/bar/baz scratch/link
+    ln -s foo/bar/nowhere scratch/broken
+    ln -s foo/bar/nowhere/nothing scratch/nested
+    ln -s foo/bar/baz/qux scratch/notdir
+    ln -s scratch/foo/bar scratch/relative
+    mkdir scratch/__bfs__
+    ln -s /__bfs__/nowhere scratch/absolute
+
+    local EXPECTED="$TESTS/${FUNCNAME[0]}.out"
+    if [ "$UPDATE" ]; then
+        local ACTUAL="$EXPECTED"
+    else
+        local ACTUAL="$TMP/${FUNCNAME[0]}.out"
+    fi
+
+    LS_COLORS="or=01;31:" invoke_bfs scratch/{,link,broken,nested,notdir,relative,absolute} -color -type l -ls \
+        | sed 's/.* -> //' \
+        | sort -o "$ACTUAL"
+
+    if [ ! "$UPDATE" ]; then
+        diff -u "$EXPECTED" "$ACTUAL"
+    fi
+}
+
 function test_deep() {
     closefrom 4
 
@@ -1926,6 +2212,22 @@ function test_de_morgan_or() {
     bfs_diff basic \( \! -name 'foo' -o \! -type f \)
 }
 
+function test_and_false_or_true() {
+    # Test (-a lhs(always_true) -false) <==> (! lhs),
+    # (-o lhs(always_false) -true) <==> (! lhs)
+    bfs_diff basic -prune -false -o -true
+}
+
+function test_comma_redundant_true() {
+    # Test (, lhs(always_true) -true) <==> lhs
+    bfs_diff basic -prune , -true
+}
+
+function test_comma_redundant_false() {
+    # Test (, lhs(always_false) -false) <==> lhs
+    bfs_diff basic -print -not -prune , -false
+}
+
 function test_data_flow_depth() {
     bfs_diff basic -depth +1 -depth -4
 }
@@ -1974,45 +2276,298 @@ function test_closed_stderr() {
     ! invoke_bfs basic >&- 2>&-
 }
 
+function test_unique() {
+    bfs_diff links/{file,symlink,hardlink} -unique
+}
+
+function test_unique_depth() {
+    bfs_diff basic -unique -depth
+}
+
+function test_L_unique() {
+    bfs_diff -L links/{file,symlink,hardlink} -unique
+}
+
+function test_L_unique_loops() {
+    bfs_diff -L loops/deeply/nested -unique
+}
+
+function test_L_unique_depth() {
+    bfs_diff -L loops/deeply/nested -unique -depth
+}
+
+function test_mount() {
+    rm -rf scratch/*
+    mkdir scratch/{foo,mnt}
+    sudo mount -t tmpfs tmpfs scratch/mnt
+    touch scratch/foo/bar scratch/mnt/baz
+
+    bfs_diff scratch -mount
+    local ret=$?
+
+    sudo umount scratch/mnt
+    return $ret
+}
+
+function test_L_mount() {
+    rm -rf scratch/*
+    mkdir scratch/{foo,mnt}
+    sudo mount -t tmpfs tmpfs scratch/mnt
+    ln -s ../mnt scratch/foo/bar
+    touch scratch/mnt/baz
+    ln -s ../mnt/baz scratch/foo/qux
+
+    bfs_diff -L scratch -mount
+    local ret=$?
+
+    sudo umount scratch/mnt
+    return $ret
+}
+
+function test_xdev() {
+    rm -rf scratch/*
+    mkdir scratch/{foo,mnt}
+    sudo mount -t tmpfs tmpfs scratch/mnt
+    touch scratch/foo/bar scratch/mnt/baz
+
+    bfs_diff scratch -xdev
+    local ret=$?
+
+    sudo umount scratch/mnt
+    return $ret
+}
+
+function test_L_xdev() {
+    rm -rf scratch/*
+    mkdir scratch/{foo,mnt}
+    sudo mount -t tmpfs tmpfs scratch/mnt
+    ln -s ../mnt scratch/foo/bar
+    touch scratch/mnt/baz
+    ln -s ../mnt/baz scratch/foo/qux
+
+    bfs_diff -L scratch -xdev
+    local ret=$?
+
+    sudo umount scratch/mnt
+    return $ret
+}
+
+function test_inum_mount() {
+    rm -rf scratch/*
+    mkdir scratch/{foo,mnt}
+    sudo mount -t tmpfs tmpfs scratch/mnt
+
+    bfs_diff scratch -inum "$(inum scratch/mnt)"
+    local ret=$?
+
+    sudo umount scratch/mnt
+    return $ret
+}
+
+function test_inum_bind_mount() {
+    rm -rf scratch/*
+    touch scratch/{foo,bar}
+    sudo mount --bind scratch/{foo,bar}
+
+    bfs_diff scratch -inum "$(inum scratch/bar)"
+    local ret=$?
+
+    sudo umount scratch/bar
+    return $ret
+}
+
+function test_type_bind_mount() {
+    rm -rf scratch/*
+    touch scratch/{file,null}
+    sudo mount --bind /dev/null scratch/null
+
+    bfs_diff scratch -type c
+    local ret=$?
+
+    sudo umount scratch/null
+    return $ret
+}
+
+function test_xtype_bind_mount() {
+    rm -rf scratch/*
+    touch scratch/{file,null}
+    sudo mount --bind /dev/null scratch/null
+    ln -s /dev/null scratch/link
+
+    bfs_diff -L scratch -type c
+    local ret=$?
+
+    sudo umount scratch/null
+    return $ret
+}
+
+function set_acl() {
+    case "$UNAME" in
+        Darwin)
+            chmod +a "$(id -un) allow read,write" "$1"
+            ;;
+        FreeBSD)
+            if [ "$(getconf ACL_NFS4 "$1")" -gt 0 ]; then
+                setfacl -m "u:$(id -un):rw::allow" "$1"
+            else
+                setfacl -m "u:$(id -un):rw" "$1"
+            fi
+            ;;
+        *)
+            setfacl -m "u:$(id -un):rw" "$1"
+            ;;
+    esac
+}
+
+function test_acl() {
+    rm -rf scratch/*
+
+    invoke_bfs scratch -quit -acl 2>/dev/null || return 0
+
+    touch scratch/{normal,acl}
+    set_acl scratch/acl || return 0
+    ln -s acl scratch/link
+
+    bfs_diff scratch -acl
+}
+
+function test_L_acl() {
+    rm -rf scratch/*
+
+    invoke_bfs scratch -quit -acl 2>/dev/null || return 0
+
+    touch scratch/{normal,acl}
+    set_acl scratch/acl || return 0
+    ln -s acl scratch/link
+
+    bfs_diff -L scratch -acl
+}
+
+function test_capable() {
+    rm -rf scratch/*
+
+    if ! invoke_bfs scratch -quit -capable 2>/dev/null; then
+        return 0
+    fi
+
+    touch scratch/{normal,capable}
+    sudo setcap all+ep scratch/capable
+    ln -s capable scratch/link
+
+    bfs_diff scratch -capable
+}
+
+function test_L_capable() {
+    rm -rf scratch/*
+
+    if ! invoke_bfs scratch -quit -capable 2>/dev/null; then
+        return 0
+    fi
+
+    touch scratch/{normal,capable}
+    sudo setcap all+ep scratch/capable
+    ln -s capable scratch/link
+
+    bfs_diff -L scratch -capable
+}
+
+function set_xattr() {
+    case "$UNAME" in
+        Darwin)
+            xattr -w bfs_test true "$1"
+            xattr -s -w bfs_test true "$2"
+            ;;
+        FreeBSD)
+            setextattr user bfs_test true "$1"
+            setextattr -h user bfs_test true "$2"
+            ;;
+        *)
+            # Linux tmpfs doesn't support the user.* namespace, so we use the security.*
+            # namespace, which is writable by root and readable by others
+            sudo setfattr -n security.bfs_test "$1"
+            sudo setfattr -h -n security.bfs_test "$2"
+            ;;
+    esac
+}
+
+function test_xattr() {
+    rm -rf scratch/*
+
+    if ! invoke_bfs scratch -quit -xattr 2>/dev/null; then
+        return 0
+    fi
+
+    touch scratch/{normal,xattr}
+    ln -s xattr scratch/link
+    ln -s normal scratch/xattr_link
+    set_xattr scratch/xattr scratch/xattr_link
+
+    bfs_diff scratch -xattr
+}
+
+function test_L_xattr() {
+    rm -rf scratch/*
+
+    if ! invoke_bfs scratch -quit -xattr 2>/dev/null; then
+        return 0
+    fi
+
+    touch scratch/{normal,xattr}
+    ln -s xattr scratch/link
+    ln -s normal scratch/xattr_link
+    set_xattr scratch/xattr scratch/xattr_link
+
+    bfs_diff -L scratch -xattr
+}
+
+function test_help() {
+    ! invoke_bfs -help | grep -E '\{...?\}'
+}
+
+
+BOL=
+EOL='\n'
+
+function update_eol() {
+    # Put the cursor at the last column, then write a space so the next
+    # character will wrap
+    EOL="\\033[${COLUMNS}G "
+}
+
 if [ -t 1 -a ! "$VERBOSE" ]; then
-    in_place=yes
+    BOL='\r\033[K'
+
+    # Workaround for bash 4: checkwinsize is off by default.  We can turn it on,
+    # but we also have to explicitly trigger a foreground job to finish so that
+    # it will update the window size before we use $COLUMNS
+    shopt -s checkwinsize
+    (:)
+
+    update_eol
+    trap update_eol WINCH
 fi
 
 passed=0
 failed=0
 
-for test in ${!run_*}; do
-    test=${test#run_}
+for test in "${enabled_tests[@]}"; do
+    printf "${BOL}${YLW}%s${RST}${EOL}" "$test"
 
-    if [ "$in_place" ]; then
-        printf '\r\033[J%s' "$test"
-    else
-        echo "$test"
-    fi
-
-    ("$test" "$dir")
+    ("$test")
     status=$?
 
     if [ $status -eq 0 ]; then
         ((++passed))
     else
         ((++failed))
-        if [ "$in_place" ]; then
-            printf '\r\033[J%s\n' "$test failed!"
-        else
-            echo "$test failed!"
-        fi
+        printf "${BOL}${RED}%s failed!${RST}\n" "$test"
     fi
 done
 
-if [ "$in_place" ]; then
-    printf '\r\033[J'
-fi
-
 if [ $passed -gt 0 ]; then
-    echo "tests passed: $passed"
+    printf "${BOL}${GRN}tests passed: %d${RST}\n" "$passed"
 fi
 if [ $failed -gt 0 ]; then
-    echo "tests failed: $failed"
+    printf "${BOL}${RED}tests failed: %s${RST}\n" "$failed"
     exit 1
 fi
